package ch.mydoli.focal.ui.photos

import android.annotation.SuppressLint
import android.graphics.Bitmap
import android.graphics.BitmapFactory
import android.graphics.Canvas
import android.graphics.Paint
import android.graphics.drawable.BitmapDrawable
import android.graphics.drawable.Drawable
import android.os.Bundle
import android.text.style.ForegroundColorSpan
import android.view.LayoutInflater
import android.view.Menu
import android.view.MenuInflater
import android.view.MenuItem
import android.view.View
import android.view.ViewGroup
import android.widget.Button
import android.widget.DatePicker
import android.widget.ImageView
import androidx.compose.ui.geometry.Offset
import androidx.compose.ui.geometry.Rect
import androidx.constraintlayout.widget.ConstraintLayout
import androidx.core.content.ContextCompat
import androidx.core.graphics.drawable.toBitmap
import androidx.core.os.bundleOf
import androidx.core.view.MenuProvider
import androidx.core.view.ViewCompat
import androidx.fragment.app.Fragment
import androidx.fragment.app.viewModels
import androidx.lifecycle.Lifecycle
import androidx.navigation.fragment.FragmentNavigatorExtras
import androidx.navigation.fragment.findNavController
import androidx.preference.PreferenceManager
import ch.mydoli.focal.R
import ch.mydoli.focal.databinding.FragmentCalendarBinding
import ch.mydoli.focal.ui.settings.SettingsFragment
import com.google.android.material.bottomnavigation.BottomNavigationView
import com.prolificinteractive.materialcalendarview.CalendarDay
import com.prolificinteractive.materialcalendarview.DayViewDecorator
import com.prolificinteractive.materialcalendarview.DayViewFacade
import com.prolificinteractive.materialcalendarview.MaterialCalendarView
import java.io.File
import java.time.LocalDate
import java.util.Timer
import kotlin.concurrent.schedule


class CalendarFragment : Fragment() {

    private var _binding: FragmentCalendarBinding? = null

    // This property is only valid between onCreateView and
    // onDestroyView.
    private val binding get() = _binding!!

    private val now = LocalDate.now()
    private var year = now.year
    private var month = now.monthValue
    private var day = now.dayOfMonth

    private var currentDate = LocalDate.of(now.year, now.month, 1)
    private var decoratedMonths: HashMap<LocalDate, HashMap<LocalDate, DayViewDecorator>> = hashMapOf()

    private lateinit var materialCalendarView: MaterialCalendarView

    private val photosViewModel: PhotosViewModel by viewModels()

    @SuppressLint("ClickableViewAccessibility")
    override fun onCreateView(
        inflater: LayoutInflater,
        container: ViewGroup?,
        savedInstanceState: Bundle?
    ): View {

        _binding = FragmentCalendarBinding.inflate(inflater, container, false)
        val root: View = binding.root

        // Fetch view elements
        materialCalendarView = binding.materialCalendarView
        val imageView: ImageView = binding.imageView
        val datePickerBackground: View = binding.datePickerBackground
        val datePickerView: DatePicker = binding.datePickerView
        val datePickerConstraint: ConstraintLayout = binding.datePickerViewConstraint
        val buttonCancel: Button = binding.buttonCancel
        val buttonOk: Button = binding.buttonOk
        val timeSpacer: View = binding.timeSpacer

        // Set first day of the week
        val sharedPreferences = PreferenceManager.getDefaultSharedPreferences(requireContext())
        SettingsFragment.setFirstDay(requireContext(), materialCalendarView, sharedPreferences.getString(getString(R.string.key_first_day), getString(R.string.sunday)))

        // Highlight initial date
        val initialDate = LocalDate.of(year, month, 1)
        // TODO: Improve efficiency (use coroutines) for startup to become entirely seamless
        addDecoratorsForMonth(initialDate)
        addDecoratorsForMonth(initialDate.minusMonths(1))
        addDecoratorsForMonth(initialDate.plusMonths(1))

        // Directly show image if any
        setYearMonthDay(year, month, day)

        // Create toolbar menu
        requireActivity().addMenuProvider(object : MenuProvider {
            override fun onCreateMenu(menu: Menu, menuInflater: MenuInflater) {
                // Add menu items here
                menuInflater.inflate(R.menu.top_calendar_menu, menu)
            }

            override fun onMenuItemSelected(menuItem: MenuItem): Boolean {
                // Handle the menu selection
                return when (menuItem.itemId) {
                    R.id.calendar_pick -> {
                        datePickerView.updateDate(year, month - 1, day)
                        datePickerConstraint.visibility = View.VISIBLE
                        datePickerBackground.visibility = View.VISIBLE
                        true
                    }
                    R.id.calendar_today -> {
                        // Button to jump (back) to today
                        val now = LocalDate.now()
                        setYearMonthDay(now.year, now.monthValue, now.dayOfMonth)
                        materialCalendarView.setDateSelected(CalendarDay.from(now.year, now.monthValue, now.dayOfMonth), true)
                        materialCalendarView.setCurrentDate(CalendarDay.from(now.year, now.monthValue, now.dayOfMonth), true)
                        materialCalendarView.selectedDate = CalendarDay.from(now.year, now.monthValue, now.dayOfMonth)
                        true
                    }
                    else -> {
                        false
                    }
                }
            }
        }, viewLifecycleOwner, Lifecycle.State.RESUMED)

        // TODO: Improve efficiency (use coroutines) for swipe to become entirely seamless
        materialCalendarView.setOnMonthChangedListener { _, calendarDay ->
            val oldDate = currentDate
            currentDate = LocalDate.of(calendarDay.year, calendarDay.month, calendarDay.day)
            val firstDayOfMonth = LocalDate.of(currentDate.year, currentDate.monthValue, 1)
            if (currentDate.plusMonths(1) != oldDate && currentDate.minusMonths(1) != oldDate) {
                // Jump anywhere
                decoratedMonths = hashMapOf()
                materialCalendarView.removeDecorators()

                addDecoratorsForMonth(currentDate)
                decoratedMonths[firstDayOfMonth]?.let {
                    it[currentDate.withDayOfMonth(day)]?.let { decorator ->
                        materialCalendarView.removeDecorator(decorator)
                        val newDecorator = EventDecorator(currentDate.withDayOfMonth(day), true)
                        it[currentDate.withDayOfMonth(day)] = newDecorator
                        materialCalendarView.addDecorator(newDecorator)
                    }
                }

                timeSpacer.visibility = View.VISIBLE
                Timer().schedule(500) {
                    requireActivity().runOnUiThread {
                        addDecoratorsForMonth(currentDate.minusMonths(1))
                        addDecoratorsForMonth(currentDate.plusMonths(1))
                        timeSpacer.visibility = View.GONE
                    }
                }
            } else {
                timeSpacer.visibility = View.VISIBLE
                Timer().schedule(500) {
                    requireActivity().runOnUiThread {
                        if (currentDate.isBefore(oldDate)) {
                            // Jump back by 1
                            val oldNextDate = oldDate.plusMonths(1)
                            val newPreviousDate = currentDate.minusMonths(1)
                            addDecoratorsForMonth(newPreviousDate)
                            removeDecoratorsForMonth(oldNextDate)
                        } else if (currentDate.isAfter(oldDate)) {
                            // Jump forward by 1
                            val oldPreviousDate = oldDate.minusMonths(1)
                            val newNextDate = currentDate.plusMonths(1)
                            addDecoratorsForMonth(newNextDate)
                            removeDecoratorsForMonth(oldPreviousDate)
                        }
                        timeSpacer.visibility = View.GONE
                    }
                }
            }
        }

        // Update current day
        materialCalendarView.setOnDateChangedListener { _, date, _ ->
            if ((this.year == date.year) && (this.month == date.month) && (this.day == date.day)) {
                // We clicked on the day that was already selected
                val bundle = bundleOf(
                    DayFragment.YEAR to year,
                    DayFragment.MONTH to month,
                    DayFragment.DAY to day,
                )
                val extras = FragmentNavigatorExtras(
                    imageView to getString(R.string.calendar_image)
                )
                findNavController().navigate(R.id.action_navigation_calendar_to_navigation_day, bundle, null, extras)
            } else {
                // We clicked on another day
                setYearMonthDay(date.year, date.month, date.day)
            }
        }

        timeSpacer.setOnTouchListener { _, _ ->
            true
        }

        datePickerBackground.setOnTouchListener { _, _ ->
            datePickerConstraint.visibility = View.GONE
            datePickerBackground.visibility = View.GONE
            true
        }

        buttonCancel.setOnClickListener {
            datePickerConstraint.visibility = View.GONE
            datePickerBackground.visibility = View.GONE
        }

        // TODO: Remove day picker to keep only month and year if possible

        buttonOk.setOnClickListener {
            val y = datePickerView.year
            val m = datePickerView.month
            val d = datePickerView.dayOfMonth
            setYearMonthDay(y, m + 1, d)
            materialCalendarView.setDateSelected(CalendarDay.from(year, month, day), true)
            materialCalendarView.setCurrentDate(CalendarDay.from(year, month, day), true)
            materialCalendarView.selectedDate = CalendarDay.from(year, month, day)
            // TODO: Change month but not date if possible
            datePickerConstraint.visibility = View.GONE
            datePickerBackground.visibility = View.GONE
        }

        imageView.setOnTouchListener { view, event ->
            val rect = Rect(view.left.toFloat(), view.top.toFloat(), view.right.toFloat(), view.bottom.toFloat())
            when (event.action) {
                android.view.MotionEvent.ACTION_DOWN -> {
                    view.alpha = 0.5F
                }
                android.view.MotionEvent.ACTION_MOVE -> {
                    if (!rect.contains(Offset(view.left + event.x, view.top + event.y))) {
                        view.alpha = 1F
                    }
                }
                android.view.MotionEvent.ACTION_UP -> {
                    view.alpha = 1F
                }
                else -> { }
            }
            false
        }

        // Click on image to open day fragment
        imageView.setOnClickListener {
            val bundle = bundleOf(
                DayFragment.YEAR to year,
                DayFragment.MONTH to month,
                DayFragment.DAY to day,
            )
            val extras = FragmentNavigatorExtras(
                imageView to getString(R.string.calendar_image)
            )
            findNavController().navigate(R.id.action_navigation_calendar_to_navigation_day, bundle, null, extras)
        }

        ViewCompat.setTransitionName(imageView, "calendar_image_view")

        return root
    }

    override fun onViewStateRestored(savedInstanceState: Bundle?) {
        super.onViewStateRestored(savedInstanceState)

        // Show bottom navigation bar
        val navView: BottomNavigationView = requireActivity().findViewById(R.id.nav_view)
        navView.visibility = View.VISIBLE

        materialCalendarView.setDateSelected(CalendarDay.from(year, month, day), true)
        materialCalendarView.setCurrentDate(CalendarDay.from(year, month, day), false)
        materialCalendarView.selectedDate = CalendarDay.from(year, month, day)
    }

    override fun onDestroyView() {
        super.onDestroyView()
        _binding = null
    }

    private fun setYearMonthDay(year: Int, month: Int, day: Int) {
        val previousFirstDayOfMonth = LocalDate.of(this.year, this.month, 1)
        val previousDate = LocalDate.of(this.year, this.month, this.day)
        decoratedMonths[previousFirstDayOfMonth]?.let {
            it[previousDate]?.let { decorator ->
                materialCalendarView.removeDecorator(decorator)
                val newDecorator = EventDecorator(previousDate, false)
                it[previousDate] = newDecorator
                materialCalendarView.addDecorator(newDecorator)
            }
        }

        // Update variables
        this.year = year
        this.month = month
        this.day = day

        val imageView = binding.imageView

        // Update view at bottom
        val date = LocalDate.of(year, month, day)
        if (photosViewModel.entryExists(date)) {
            val imageName = photosViewModel.getImage(date)
            val imgFile =
                File(requireActivity().filesDir.toString() + "/" + date.toString() + "_" + imageName)
            if (imgFile.exists()) {
                val imgBitmap = BitmapFactory.decodeFile(imgFile.absolutePath)
                imageView.setImageBitmap(imgBitmap)
            } else {
                imageView.setImageDrawable(ContextCompat.getDrawable(requireContext(), R.drawable.ic_empty_image))
            }
        } else {
            imageView.setImageDrawable(ContextCompat.getDrawable(requireContext(), R.drawable.ic_empty_image))
        }

        val firstDayOfMonth = LocalDate.of(year, month, 1)
        decoratedMonths[firstDayOfMonth]?.let {
            it[date]?.let { decorator ->
                materialCalendarView.removeDecorator(decorator)
                val newDecorator = EventDecorator(date, true)
                it[date] = newDecorator
                materialCalendarView.addDecorator(newDecorator)
            }
        }


    }

    private fun addDecoratorsForMonth(localDate: LocalDate) {
        if (localDate.dayOfMonth != 1) {
            throw IllegalArgumentException()
        }
        for (i in 1.rangeTo(localDate.lengthOfMonth())) {
            val iDate = LocalDate.of(localDate.year, localDate.month, i)
            val iDecorator = EventDecorator(iDate, iDate == LocalDate.of(year, month, day))
            if (!decoratedMonths.containsKey(localDate)) {
                decoratedMonths[localDate] = hashMapOf(Pair(iDate, iDecorator))
            } else {
                decoratedMonths[localDate]!![iDate] = iDecorator
            }
            materialCalendarView.addDecorator(iDecorator)
        }
    }

    private fun removeDecoratorsForMonth(localDate: LocalDate) {
        if (localDate.dayOfMonth != 1) {
            throw IllegalArgumentException()
        }
        if (decoratedMonths.containsKey(localDate)) {
            for ((_, decorator) in decoratedMonths[localDate]!!) {
                materialCalendarView.removeDecorator(decorator)
            }
            decoratedMonths.remove(localDate)
        }
    }

    private inner class EventDecorator(date: LocalDate, val selected: Boolean):
        DayViewDecorator {
        private val drawable: Drawable
        private val calendarDay: CalendarDay

        init {
            this.calendarDay = CalendarDay.from(date.year, date.monthValue, date.dayOfMonth)
            val thumbBitmap = if (photosViewModel.entryExists(date)) {
                val imageName = photosViewModel.getImage(date)
                val thumbFile =
                    File(requireActivity().filesDir.toString() + "/" + date.toString() + "_" + imageName + "_thumb")
                if (thumbFile.exists()) {
                    val imgBitmap = BitmapFactory.decodeFile(thumbFile.absolutePath)
                    val smallBitmap = Bitmap.createScaledBitmap(imgBitmap, 150, 150, true)
                    smallBitmap
                } else {
                    val resDrawable = ContextCompat.getDrawable(requireContext(), R.drawable.ic_empty_image)!!
                    resDrawable.alpha = 127
                    resDrawable.toBitmap(150, 150)
                }
            } else {
                val resDrawable = ContextCompat.getDrawable(requireContext(), R.drawable.ic_empty_image)!!
                resDrawable.alpha = 127
                resDrawable.toBitmap(150, 150)
            }

            val borderSize = thumbBitmap.width / 10
            val bitmap = Bitmap.createBitmap(thumbBitmap.width + borderSize * 2, thumbBitmap.height + borderSize * 2, thumbBitmap.config)

            val canvas = Canvas(bitmap)
            canvas.drawBitmap(thumbBitmap, borderSize + 0.0f, borderSize + 0.0f, Paint().apply {
                alpha = if (selected) {
                    127
                } else {
                    255
                }
            })

            this.drawable = BitmapDrawable(resources, bitmap)
        }

        override fun shouldDecorate(calendarDay: CalendarDay): Boolean {
            //val year = calendarDay.year
            //val month = calendarDay.month
            //val day = calendarDay.day
            return this.calendarDay == calendarDay// && photosViewModel.entryExists(LocalDate.of(year, month, day))
        }

        override fun decorate(view: DayViewFacade) {
            view.setBackgroundDrawable(drawable)
            view.setSelectionDrawable(drawable)
            view.addSpan(ForegroundColorSpan(ContextCompat.getColor(requireContext(), R.color.white)))
        }
    }
}