package ch.mydoli.focal.ui.photos

import android.content.Context
import android.content.Intent
import android.graphics.Bitmap
import android.graphics.BitmapFactory
import android.graphics.Color
import android.net.Uri
import android.os.Bundle
import android.text.SpannableStringBuilder
import android.transition.TransitionInflater
import android.view.LayoutInflater
import android.view.Menu
import android.view.MenuInflater
import android.view.MenuItem
import android.view.View
import android.view.ViewGroup
import android.view.inputmethod.InputMethodManager
import android.widget.EditText
import android.widget.ImageView
import androidx.activity.OnBackPressedCallback
import androidx.appcompat.app.AppCompatActivity
import androidx.core.content.ContextCompat
import androidx.core.os.bundleOf
import androidx.core.view.MenuProvider
import androidx.fragment.app.Fragment
import androidx.fragment.app.viewModels
import androidx.lifecycle.Lifecycle
import androidx.navigation.fragment.FragmentNavigatorExtras
import androidx.navigation.fragment.findNavController
import androidx.preference.PreferenceManager
import ch.mydoli.focal.R
import ch.mydoli.focal.databinding.FragmentDayBinding
import ch.mydoli.focal.ui.settings.SettingsFragment
import com.canhub.cropper.CropImageContract
import com.canhub.cropper.CropImageContractOptions
import com.canhub.cropper.CropImageOptions
import com.google.android.material.bottomnavigation.BottomNavigationView
import com.google.android.material.dialog.MaterialAlertDialogBuilder
import java.io.File
import java.io.FileOutputStream
import java.time.LocalDate


class DayFragment : Fragment() {

    companion object {
        internal const val YEAR = "year"
        internal const val MONTH = "month"
        internal const val DAY = "day"
    }

    private var _binding: FragmentDayBinding? = null

    // This property is only valid between onCreateView and
    // onDestroyView.
    private val binding get() = _binding!!

    private lateinit var topMenu: Menu
    private lateinit var favoriteMenuItem: MenuItem
    private lateinit var deleteMenuItem: MenuItem
    private lateinit var saveMenuItem: MenuItem
    private lateinit var editMenuItem: MenuItem
    private lateinit var cancelMenuItem: MenuItem

    private var editMode = false
    private var oldTitle = ""
    private var oldDescription = ""
    private var oldImageName = ""
    private var editedTitle = ""
    private var editedDescription = ""

    private val newImages = mutableListOf<String>()

    private lateinit var date: LocalDate

    private lateinit var textTitle: EditText
    private lateinit var textDescription: EditText
    private lateinit var imageDay: ImageView

    private val photosViewModel: PhotosViewModel by viewModels()

    override fun onCreateView(
            inflater: LayoutInflater,
            container: ViewGroup?,
            savedInstanceState: Bundle?
    ): View {

        _binding = FragmentDayBinding.inflate(inflater, container, false)
        val root: View = binding.root

        // Hide bottom navigation bar
        val navView: BottomNavigationView = requireActivity().findViewById(R.id.nav_view)
        navView.visibility = View.GONE

        val now = LocalDate.now()

        val year = arguments?.getInt(YEAR)?: now.year
        val month = arguments?.getInt(MONTH)?: now.monthValue
        val day = arguments?.getInt(DAY)?: now.dayOfMonth

        textTitle = binding.textTitle
        textDescription = binding.textDescription
        imageDay = binding.imageDay

        // Disable text edition by default
        disableEditText(textTitle)
        disableEditText(textDescription)

        // Write date into title label
        date = LocalDate.of(year, month, day)

        // Get context and shared preferences
        val context = requireContext()
        val sharedPreferences = PreferenceManager.getDefaultSharedPreferences(context)

        // Format date
        (requireActivity() as AppCompatActivity).supportActionBar?.title = SettingsFragment.getDateFormat(context, date, sharedPreferences.getString(context.getString(R.string.key_date_format), context.getString(R.string.yyyymmdd)))

        // Create toolbar menu
        requireActivity().addMenuProvider(object : MenuProvider {
            override fun onCreateMenu(menu: Menu, menuInflater: MenuInflater) {
                // Add menu items here
                menuInflater.inflate(R.menu.top_feed_menu, menu)
                topMenu = menu
                favoriteMenuItem = menu.findItem(R.id.feed_favorite)
                deleteMenuItem = menu.findItem(R.id.feed_delete)
                saveMenuItem = menu.findItem(R.id.feed_save)
                editMenuItem = menu.findItem(R.id.feed_edit)
                cancelMenuItem = menu.findItem(R.id.feed_cancel)

                // Choose which favorite button to display
                if (photosViewModel.entryExists(date)) {
                    favoriteMenuItem.isVisible = true
                    deleteMenuItem.isVisible = true
                    if (photosViewModel.isFavorite(date)) {
                        favoriteMenuItem.setIcon(R.drawable.ic_favorites_black_24dp)
                    } else {
                        favoriteMenuItem.setIcon(R.drawable.ic_notfavorites_black_24dp)
                    }

                    // Reconstruct menu if coming back from image picker
                    if (editMode) {
                        editMenuItem.isVisible = false
                        favoriteMenuItem.isVisible = false
                        deleteMenuItem.isVisible = false
                        saveMenuItem.isVisible = true
                        cancelMenuItem.isVisible = true
                    }
                } else {
                    favoriteMenuItem.isVisible = false
                    deleteMenuItem.isVisible = false
                }
            }

            override fun onMenuItemSelected(menuItem: MenuItem): Boolean {
                // Handle the menu selection
                return when (menuItem.itemId) {
                    android.R.id.home -> {
                        if (editMode && ((photosViewModel.entryExists(date) && oldImageName != photosViewModel.getImage(date)) || (oldTitle != textTitle.text.toString()) || (oldDescription != textTitle.text.toString()) || newImages.isNotEmpty())) {
                            MaterialAlertDialogBuilder(requireContext())
                                .setMessage(R.string.dialog_back)
                                .setPositiveButton(R.string.dialog_back_cancel) { _, _ -> }
                                .setNegativeButton(R.string.dialog_back_save) { _, _ ->
                                    disableEditText(textTitle)
                                    disableEditText(textDescription)
                                    editMode = false
                                    photosViewModel.changeTitleDescription(date, textTitle.text.toString(), textDescription.text.toString())

                                    if (newImages.isNotEmpty()) {
                                        val newImageName = newImages.removeLast()

                                        // Delete old image file
                                        newImages.add(oldImageName)

                                        // Delete unused intermediary image files
                                        newImages.forEach {
                                            val oldImgFile =
                                                File(requireActivity().filesDir.toString() + "/" + date.toString() + "_" + it)
                                            val oldThumbFile =
                                                File(requireActivity().filesDir.toString() + "/" + date.toString() + "_" + it + "_thumb")
                                            if (oldImgFile.exists() && it != newImageName) {
                                                oldImgFile.delete()
                                                oldThumbFile.delete()
                                            }
                                        }

                                        newImages.removeAll(newImages)
                                    }
                                    oldImageName = ""
                                    oldTitle = ""
                                    oldDescription = ""
                                    editMenuItem.isVisible = true
                                    val entryExists = photosViewModel.entryExists(date)
                                    favoriteMenuItem.isVisible = entryExists
                                    deleteMenuItem.isVisible = entryExists
                                    saveMenuItem.isVisible = false
                                    cancelMenuItem.isVisible = false
                                    if (entryExists && textTitle.text.toString() == "" && textDescription.text.toString() == "" && photosViewModel.getImage(date) == "") {
                                        photosViewModel.deleteEntry(date)
                                    }
                                    reloadFragment()
                                    findNavController().navigateUp()
                                }
                                .setNeutralButton(R.string.dialog_back_discard) { _, _ ->
                                    if (photosViewModel.getImage(date) != oldImageName || textTitle.text.toString() != oldTitle || textDescription.text.toString() != oldDescription) {
                                        disableEditText(textTitle)
                                        disableEditText(textDescription)
                                        editMode = false
                                        textTitle.text = SpannableStringBuilder(oldTitle)
                                        textDescription.text = SpannableStringBuilder(oldDescription)

                                        if (newImages.isNotEmpty()) {
                                            // Delete unused intermediary image files
                                            newImages.forEach {
                                                val oldImgFile =
                                                    File(requireActivity().filesDir.toString() + "/" + date.toString() + "_" + it)
                                                val oldThumbFile =
                                                    File(requireActivity().filesDir.toString() + "/" + date.toString() + "_" + it + "_thumb")
                                                if (oldImgFile.exists() && it != oldImageName) {
                                                    oldImgFile.delete()
                                                    oldThumbFile.delete()
                                                }
                                            }

                                            newImages.removeAll(newImages)

                                            if (oldImageName == "" && oldTitle == "" && oldDescription == "") {
                                                photosViewModel.deleteEntry(date)
                                            } else {
                                                photosViewModel.changeImage(date, oldImageName)
                                            }
                                        }

                                        editMenuItem.isVisible = true
                                        val entryExists = photosViewModel.entryExists(date)
                                        favoriteMenuItem.isVisible = entryExists
                                        deleteMenuItem.isVisible = entryExists
                                        saveMenuItem.isVisible = false
                                        cancelMenuItem.isVisible = false
                                        oldTitle = ""
                                        oldDescription = ""
                                        oldImageName = ""
                                        reloadFragment()
                                    } else {
                                        disableEditText(textTitle)
                                        disableEditText(textDescription)
                                        editMode = false

                                        if (newImages.isNotEmpty()) {
                                            // Delete unused intermediary image files
                                            newImages.forEach {
                                                val oldImgFile =
                                                    File(requireActivity().filesDir.toString() + "/" + date.toString() + "_" + it)
                                                val oldThumbFile =
                                                    File(requireActivity().filesDir.toString() + "/" + date.toString() + "_" + it + "_thumb")
                                                if (oldImgFile.exists() && it != oldImageName) {
                                                    oldImgFile.delete()
                                                    oldThumbFile.delete()
                                                }
                                            }

                                            newImages.removeAll(newImages)

                                            if (oldImageName == "" && oldTitle == "" && oldDescription == "") {
                                                photosViewModel.deleteEntry(date)
                                            }
                                        }

                                        editMenuItem.isVisible = true
                                        val entryExists = photosViewModel.entryExists(date)
                                        favoriteMenuItem.isVisible = entryExists
                                        deleteMenuItem.isVisible = entryExists
                                        saveMenuItem.isVisible = false
                                        cancelMenuItem.isVisible = false
                                        oldTitle = ""
                                        oldDescription = ""
                                        oldImageName = ""
                                        reloadFragment()
                                    }
                                    findNavController().navigateUp()
                                }.show()
                        } else {
                            findNavController().navigateUp()
                        }
                        true
                    }
                    R.id.feed_cancel -> {
                        if (photosViewModel.entryExists(date) && (photosViewModel.getImage(date) != oldImageName || textTitle.text.toString() != oldTitle || textDescription.text.toString() != oldDescription)) {
                            MaterialAlertDialogBuilder(requireContext())
                                .setMessage(R.string.dialog_cancel)
                                .setPositiveButton(R.string.dialog_continue_editing) { _, _ -> }
                                .setNegativeButton(R.string.dialog_cancel_edits) { _, _ ->
                                    disableEditText(textTitle)
                                    disableEditText(textDescription)
                                    editMode = false
                                    textTitle.text = SpannableStringBuilder(oldTitle)
                                    textDescription.text = SpannableStringBuilder(oldDescription)

                                    if (newImages.isNotEmpty()) {
                                        // Delete unused intermediary image files
                                        newImages.forEach {
                                            val oldImgFile =
                                                File(requireActivity().filesDir.toString() + "/" + date.toString() + "_" + it)
                                            val oldThumbFile =
                                                File(requireActivity().filesDir.toString() + "/" + date.toString() + "_" + it + "_thumb")
                                            if (oldImgFile.exists() && it != oldImageName) {
                                                oldImgFile.delete()
                                                oldThumbFile.delete()
                                            }
                                        }

                                        newImages.removeAll(newImages)

                                        if (oldImageName == "" && oldTitle == "" && oldDescription == "") {
                                            photosViewModel.deleteEntry(date)
                                        } else {
                                            photosViewModel.changeImage(date, oldImageName)
                                        }
                                    }

                                    editMenuItem.isVisible = true
                                    val entryExists = photosViewModel.entryExists(date)
                                    favoriteMenuItem.isVisible = entryExists
                                    deleteMenuItem.isVisible = entryExists
                                    saveMenuItem.isVisible = false
                                    cancelMenuItem.isVisible = false
                                    oldTitle = ""
                                    oldDescription = ""
                                    oldImageName = ""
                                    reloadFragment()
                                }.show()
                        } else {
                            disableEditText(textTitle)
                            disableEditText(textDescription)
                            editMode = false

                            if (newImages.isNotEmpty()) {
                                // Delete unused intermediary image files
                                newImages.forEach {
                                    val oldImgFile =
                                        File(requireActivity().filesDir.toString() + "/" + date.toString() + "_" + it)
                                    val oldThumbFile =
                                        File(requireActivity().filesDir.toString() + "/" + date.toString() + "_" + it + "_thumb")
                                    if (oldImgFile.exists() && it != oldImageName) {
                                        oldImgFile.delete()
                                        oldThumbFile.delete()
                                    }
                                }

                                newImages.removeAll(newImages)

                                if (oldImageName == "" && oldTitle == "" && oldDescription == "") {
                                    photosViewModel.deleteEntry(date)
                                }
                            }

                            editMenuItem.isVisible = true
                            val entryExists = photosViewModel.entryExists(date)
                            favoriteMenuItem.isVisible = entryExists
                            deleteMenuItem.isVisible = entryExists
                            saveMenuItem.isVisible = false
                            cancelMenuItem.isVisible = false
                            oldTitle = ""
                            oldDescription = ""
                            oldImageName = ""
                            reloadFragment()
                        }
                        true
                    }
                    R.id.feed_save -> {
                        disableEditText(textTitle)
                        disableEditText(textDescription)
                        editMode = false
                        photosViewModel.changeTitleDescription(date, textTitle.text.toString(), textDescription.text.toString())

                        if (newImages.isNotEmpty()) {
                            val newImageName = newImages.removeLast()

                            // Delete old image file
                            newImages.add(oldImageName)

                            // Delete unused intermediary image files
                            newImages.forEach {
                                val oldImgFile =
                                    File(requireActivity().filesDir.toString() + "/" + date.toString() + "_" + it)
                                val oldThumbFile =
                                    File(requireActivity().filesDir.toString() + "/" + date.toString() + "_" + it + "_thumb")
                                if (oldImgFile.exists() && it != newImageName) {
                                    oldImgFile.delete()
                                    oldThumbFile.delete()
                                }
                            }

                            newImages.removeAll(newImages)
                        }
                        oldImageName = ""
                        oldTitle = ""
                        oldDescription = ""
                        editMenuItem.isVisible = true
                        val entryExists = photosViewModel.entryExists(date)
                        favoriteMenuItem.isVisible = entryExists
                        deleteMenuItem.isVisible = entryExists
                        saveMenuItem.isVisible = false
                        cancelMenuItem.isVisible = false
                        if (entryExists && textTitle.text.toString() == "" && textDescription.text.toString() == "" && photosViewModel.getImage(date) == "") {
                            photosViewModel.deleteEntry(date)
                        }
                        reloadFragment()
                        true
                    }
                    R.id.feed_edit -> {
                        enableEditText(textTitle)
                        enableEditText(textDescription)
                        editMode = true
                        oldTitle = textTitle.text.toString()
                        oldDescription = textDescription.text.toString()
                        oldImageName = if (photosViewModel.entryExists(date)) {
                            photosViewModel.getImage(date)
                        } else {
                            ""
                        }
                        editMenuItem.isVisible = false
                        favoriteMenuItem.isVisible = false
                        deleteMenuItem.isVisible = false
                        saveMenuItem.isVisible = true
                        cancelMenuItem.isVisible = true
                        true
                    }
                    R.id.feed_favorite -> {
                        if (photosViewModel.isFavorite(date)) {
                            photosViewModel.unsetFavorite(date)
                            menuItem.setIcon(R.drawable.ic_notfavorites_black_24dp)
                        } else {
                            photosViewModel.setFavorite(date)
                            menuItem.setIcon(R.drawable.ic_favorites_black_24dp)
                        }
                        true
                    }
                    R.id.feed_delete -> {
                        MaterialAlertDialogBuilder(requireContext())
                            .setMessage(R.string.dialog_delete)
                            .setPositiveButton(R.string.dialog_delete_cancel) { _, _ -> }
                            .setNegativeButton(R.string.dialog_delete_image) { _, _ ->
                                val imageName = photosViewModel.getImage(date)
                                photosViewModel.changeImage(date, "")

                                // Delete image file
                                val imgFile =
                                    File(requireActivity().filesDir.toString() + "/" + date.toString() + "_" + imageName)
                                val thumbFile =
                                    File(requireActivity().filesDir.toString() + "/" + date.toString() + "_" + imageName + "_thumb")
                                if (imgFile.exists()) {
                                    imgFile.delete()
                                }
                                if (thumbFile.exists()) {
                                    thumbFile.delete()
                                }

                                // Delete entire entry if nothing else than image
                                if (textTitle.text.toString() == "" && textDescription.text.toString() == "") {
                                    photosViewModel.deleteEntry(date)
                                }

                                // Reload fragment
                                reloadFragment()
                            }
                            .setNeutralButton(R.string.dialog_delete_day) { _, _ ->
                                val imageName = photosViewModel.getImage(date)
                                photosViewModel.deleteEntry(date)

                                // Delete image file
                                val imgFile =
                                    File(requireActivity().filesDir.toString() + "/" + date.toString() + "_" + imageName)
                                val thumbFile =
                                    File(requireActivity().filesDir.toString() + "/" + date.toString() + "_" + imageName + "_thumb")
                                if (imgFile.exists()) {
                                    imgFile.delete()
                                }
                                if (thumbFile.exists()) {
                                    thumbFile.delete()
                                }

                                // Reload fragment
                                reloadFragment()
                            }
                            .show()
                        true
                    }
                    else -> {
                        false
                    }
                }
            }
        }, viewLifecycleOwner, Lifecycle.State.RESUMED)

        val cropImageResultContract = registerForActivityResult(CropImageContract()) { result ->
            if (result.isSuccessful) {
                val image = result.uriContent.toString()
                if (image != "") {
                    val imageUri = Uri.parse(image)
                    val imageName = imageUri.lastPathSegment!!

                    val imageFile =
                        File(requireActivity().filesDir.toString() + "/" + date.toString() + "_" + imageName)
                    val thumbnailFile =
                        File(requireActivity().filesDir.toString() + "/" + date.toString() + "_" + imageName + "_thumb")

                    requireActivity().contentResolver.openInputStream(imageUri).use {
                        // Apply acceptable compression (by default 50% seems good enough, but this is customizable to 25% or 75%)
                        val quality = SettingsFragment.getQuality(context, sharedPreferences.getString(context.getString(R.string.key_quality), context.getString(R.string.quality_compromise)))

                        val bitmap = BitmapFactory.decodeStream(it)

                        val scaledBitmap = Bitmap.createScaledBitmap(bitmap, 1500, 1500, true)
                        scaledBitmap.compress(Bitmap.CompressFormat.JPEG, quality, FileOutputStream(imageFile))

                        val thumbmap = Bitmap.createScaledBitmap(bitmap, 300, 300, true)
                        thumbmap.compress(Bitmap.CompressFormat.JPEG, quality, FileOutputStream(thumbnailFile))
                    }

                    newImages.add(imageName)
                    if (photosViewModel.entryExists(date)) {
                        photosViewModel.changeImage(date, imageName)
                    } else {
                        photosViewModel.setNewEntry(date, "", "", imageName)
                    }

                    // Reload fragment
                    reloadFragment()
                }
            }
        }

        textTitle.setOnClickListener {
            if (!editMode && (!photosViewModel.entryExists(date) || photosViewModel.getTitle(date) == "")) {
                enableEditText(textTitle)
                enableEditText(textDescription)
                editMode = true
                oldTitle = textTitle.text.toString()
                oldDescription = textDescription.text.toString()
                oldImageName = if (photosViewModel.entryExists(date)) {
                    photosViewModel.getImage(date)
                } else {
                    ""
                }
                editMenuItem.isVisible = false
                favoriteMenuItem.isVisible = false
                deleteMenuItem.isVisible = false
                saveMenuItem.isVisible = true
                cancelMenuItem.isVisible = true
                textTitle.requestFocus()
                val imm = requireActivity().getSystemService(Context.INPUT_METHOD_SERVICE) as InputMethodManager?
                imm!!.showSoftInput(textTitle, InputMethodManager.SHOW_IMPLICIT)
            }
        }

        textDescription.setOnClickListener {
            if (!editMode && (!photosViewModel.entryExists(date) || photosViewModel.getDescription(date) == "")) {
                enableEditText(textTitle)
                enableEditText(textDescription)
                editMode = true
                oldTitle = textTitle.text.toString()
                oldDescription = textDescription.text.toString()
                oldImageName = if (photosViewModel.entryExists(date)) {
                    photosViewModel.getImage(date)
                } else {
                    ""
                }
                editMenuItem.isVisible = false
                favoriteMenuItem.isVisible = false
                deleteMenuItem.isVisible = false
                saveMenuItem.isVisible = true
                cancelMenuItem.isVisible = true
                textDescription.requestFocus()
                val imm = requireActivity().getSystemService(Context.INPUT_METHOD_SERVICE) as InputMethodManager?
                imm!!.showSoftInput(textDescription, InputMethodManager.SHOW_IMPLICIT)
            }
        }

        imageDay.setOnClickListener {
            if (!photosViewModel.entryExists(date) || photosViewModel.getImage(date) == "" || editMode) {
                if (!editMode) {
                    oldImageName = if (photosViewModel.entryExists(date)) {
                        photosViewModel.getImage(date)
                    } else {
                        ""
                    }
                    oldTitle = textTitle.text.toString()
                    oldDescription = textDescription.text.toString()
                    enableEditText(textTitle)
                    enableEditText(textDescription)
                    editMode = true
                    editMenuItem.isVisible = false
                    favoriteMenuItem.isVisible = false
                    deleteMenuItem.isVisible = false
                    saveMenuItem.isVisible = true
                    cancelMenuItem.isVisible = true
                }
                editedTitle = textTitle.text.toString()
                editedDescription = textDescription.text.toString()

                val cameraEnabled = sharedPreferences.getBoolean(getString(R.string.key_camera), false)

                Intent(Intent.ACTION_GET_CONTENT).apply {
                    type = "image/*"
                    addCategory(Intent.CATEGORY_OPENABLE)
                    val options = CropImageOptions()
                    options.activityBackgroundColor =
                        Color.BLACK // TODO: Change according to dark/light theme
                    options.cropMenuCropButtonIcon = R.drawable.ic_done_black_24dp
                    options.showCropLabel = true
                    options.outputRequestWidth = 400
                    options.outputRequestHeight = 400
                    options.aspectRatioX = 400
                    options.aspectRatioY = 400
                    options.fixAspectRatio = true
                    options.imageSourceIncludeGallery = true
                    options.imageSourceIncludeCamera = cameraEnabled
                    options.outputCompressFormat = Bitmap.CompressFormat.JPEG
                    cropImageResultContract.launch(
                        CropImageContractOptions(
                            null,
                            options
                        )
                    )
                }

            } else {
                // Navigate to a full screen view with imageView only
                val bundle = bundleOf(
                    PhotoFragment.IMAGE to photosViewModel.getImage(date),
                    PhotoFragment.DATE to date.toString()
                )
                val extras = FragmentNavigatorExtras(
                    imageDay to getString(R.string.day_image)
                )
                imageDay.alpha = 0.5F
                findNavController().navigate(R.id.action_navigation_day_to_navigation_photo, bundle, null, extras)
            }
        }

        loadFragment()

        return root
    }

    // TODO: Make transition work
    override fun onCreate(savedInstanceState: Bundle?) {
        super.onCreate(savedInstanceState)
        sharedElementEnterTransition = TransitionInflater.from(requireContext())
            .inflateTransition(R.transition.shared_image)
    }

    override fun onViewCreated(view: View, savedInstanceState: Bundle?) {
        super.onViewCreated(view, savedInstanceState)
        // Handle back pressed button
        requireActivity().onBackPressedDispatcher.addCallback(viewLifecycleOwner, object: OnBackPressedCallback(true) {
            override fun handleOnBackPressed() {
                if (editMode && ((photosViewModel.entryExists(date) && oldImageName != photosViewModel.getImage(date)) || (oldTitle != textTitle.text.toString()) || (oldDescription != textTitle.text.toString()) || newImages.isNotEmpty())) {
                    MaterialAlertDialogBuilder(requireContext())
                        .setMessage(R.string.dialog_back)
                        .setPositiveButton(R.string.dialog_back_cancel) { _, _ -> }
                        .setNegativeButton(R.string.dialog_back_save) { _, _ ->
                            topMenu.performIdentifierAction(R.id.feed_save, 0)
                            disableEditText(textTitle)
                            disableEditText(textDescription)
                            editMode = false
                            photosViewModel.changeTitleDescription(date, textTitle.text.toString(), textDescription.text.toString())

                            if (newImages.isNotEmpty()) {
                                val newImageName = newImages.removeLast()

                                // Delete old image file
                                newImages.add(oldImageName)

                                // Delete unused intermediary image files
                                newImages.forEach {
                                    val oldImgFile =
                                        File(requireActivity().filesDir.toString() + "/" + date.toString() + "_" + it)
                                    val oldThumbFile =
                                        File(requireActivity().filesDir.toString() + "/" + date.toString() + "_" + it + "_thumb")
                                    if (oldImgFile.exists() && it != newImageName) {
                                        oldImgFile.delete()
                                        oldThumbFile.delete()
                                    }
                                }

                                newImages.removeAll(newImages)
                            }
                            oldImageName = ""
                            oldTitle = ""
                            oldDescription = ""
                            editMenuItem.isVisible = true
                            val entryExists = photosViewModel.entryExists(date)
                            favoriteMenuItem.isVisible = entryExists
                            deleteMenuItem.isVisible = entryExists
                            saveMenuItem.isVisible = false
                            cancelMenuItem.isVisible = false
                            if (entryExists && textTitle.text.toString() == "" && textDescription.text.toString() == "" && photosViewModel.getImage(date) == "") {
                                photosViewModel.deleteEntry(date)
                            }
                            reloadFragment()
                        }
                        .setNeutralButton(R.string.dialog_back_discard) { _, _ ->
                            if (photosViewModel.getImage(date) != oldImageName || textTitle.text.toString() != oldTitle || textDescription.text.toString() != oldDescription) {
                                disableEditText(textTitle)
                                disableEditText(textDescription)
                                editMode = false
                                textTitle.text = SpannableStringBuilder(oldTitle)
                                textDescription.text = SpannableStringBuilder(oldDescription)

                                if (newImages.isNotEmpty()) {
                                    // Delete unused intermediary image files
                                    newImages.forEach {
                                        val oldImgFile =
                                            File(requireActivity().filesDir.toString() + "/" + date.toString() + "_" + it)
                                        val oldThumbFile =
                                            File(requireActivity().filesDir.toString() + "/" + date.toString() + "_" + it + "_thumb")
                                        if (oldImgFile.exists() && it != oldImageName) {
                                            oldImgFile.delete()
                                            oldThumbFile.delete()
                                        }
                                    }

                                    newImages.removeAll(newImages)

                                    if (oldImageName == "" && oldTitle == "" && oldDescription == "") {
                                        photosViewModel.deleteEntry(date)
                                    } else {
                                        photosViewModel.changeImage(date, oldImageName)
                                    }
                                }

                                editMenuItem.isVisible = true
                                val entryExists = photosViewModel.entryExists(date)
                                favoriteMenuItem.isVisible = entryExists
                                deleteMenuItem.isVisible = entryExists
                                saveMenuItem.isVisible = false
                                cancelMenuItem.isVisible = false
                                oldTitle = ""
                                oldDescription = ""
                                oldImageName = ""
                                reloadFragment()
                            } else {
                                disableEditText(textTitle)
                                disableEditText(textDescription)
                                editMode = false

                                if (newImages.isNotEmpty()) {
                                    // Delete unused intermediary image files
                                    newImages.forEach {
                                        val oldImgFile =
                                            File(requireActivity().filesDir.toString() + "/" + date.toString() + "_" + it)
                                        val oldThumbFile =
                                            File(requireActivity().filesDir.toString() + "/" + date.toString() + "_" + it + "_thumb")
                                        if (oldImgFile.exists() && it != oldImageName) {
                                            oldImgFile.delete()
                                            oldThumbFile.delete()
                                        }
                                    }

                                    newImages.removeAll(newImages)

                                    if (oldImageName == "" && oldTitle == "" && oldDescription == "") {
                                        photosViewModel.deleteEntry(date)
                                    }
                                }

                                editMenuItem.isVisible = true
                                val entryExists = photosViewModel.entryExists(date)
                                favoriteMenuItem.isVisible = entryExists
                                deleteMenuItem.isVisible = entryExists
                                saveMenuItem.isVisible = false
                                cancelMenuItem.isVisible = false
                                oldTitle = ""
                                oldDescription = ""
                                oldImageName = ""
                                reloadFragment()
                            }
                            findNavController().navigateUp()
                        }.show()
                } else {
                    findNavController().navigateUp()
                }
            }
        })
    }



    override fun onViewStateRestored(savedInstanceState: Bundle?) {
        super.onViewStateRestored(savedInstanceState)
        if (editMode) {
            editMenuItem.isVisible = false
            favoriteMenuItem.isVisible = false
            deleteMenuItem.isVisible = false
            saveMenuItem.isVisible = true
            cancelMenuItem.isVisible = true
        }
    }

    override fun onDestroyView() {
        super.onDestroyView()
        _binding = null
    }

    private fun loadFragment() {
        if (photosViewModel.entryExists(date)) {
            textTitle.text = SpannableStringBuilder(photosViewModel.getTitle(date))
            textDescription.text = SpannableStringBuilder(photosViewModel.getDescription(date))

            val imageName = photosViewModel.getImage(date)
            val imgFile = File(requireActivity().filesDir.toString() + "/" + date.toString() + "_" + imageName)
            if (imgFile.exists()) {
                val imgBitmap = BitmapFactory.decodeFile(imgFile.absolutePath)
                imageDay.setImageBitmap(imgBitmap)
            } else {
                imageDay.setImageDrawable(ContextCompat.getDrawable(requireContext(), R.drawable.ic_empty_image))
            }
        } else {
            textTitle.text = SpannableStringBuilder("")
            textDescription.text = SpannableStringBuilder("")
            imageDay.setImageDrawable(ContextCompat.getDrawable(requireContext(), R.drawable.ic_empty_image))
        }
    }

    fun reloadFragment() {
        if (photosViewModel.entryExists(date)) {
            favoriteMenuItem.isVisible = true
            deleteMenuItem.isVisible = true

            if (photosViewModel.isFavorite(date)) {
                favoriteMenuItem.setIcon(R.drawable.ic_favorites_black_24dp)
            } else {
                favoriteMenuItem.setIcon(R.drawable.ic_notfavorites_black_24dp)
            }
        } else {
            favoriteMenuItem.isVisible = false
            deleteMenuItem.isVisible = false
        }

        loadFragment()

        if (editMode) {
            textTitle.text = SpannableStringBuilder(editedTitle)
            textDescription.text = SpannableStringBuilder(editedDescription)
            editedTitle = ""
            editedDescription = ""
        }
    }
    
    private fun disableEditText(editText: EditText) {
        editText.isFocusable = false
        editText.isLongClickable = false
        editText.isCursorVisible = false
        editText.isEnabled = false // disable to remove keyboard and autocorrect underlining
        editText.isEnabled = true // re-enable to preserve scrolling
        editText.clearFocus()
        editText.clearAnimation()
        editText.alpha = 1F
    }

    private fun enableEditText(editText: EditText) {
        editText.isFocusableInTouchMode = true
        editText.isLongClickable = true
        editText.isCursorVisible = true
        editText.isEnabled = true
        editText.alpha = 0.75F
    }

}