package ch.mydoli.focal.ui.photos

import android.app.Application
import android.content.ContentValues
import androidx.lifecycle.AndroidViewModel
import ch.mydoli.focal.sql.database.FeedContract
import java.lang.IllegalStateException
import java.time.LocalDate

class PhotosViewModel(application: Application) : AndroidViewModel(application) {

    companion object {
        class DBEntry(val title: String, val description: String, val imageUri: String, val favorite: Boolean = false)
    }

    // Access the database
    private val dbHelper = FeedContract.FeedDbHelper(application)

    private fun setNewEntry(date: LocalDate, dbEntry: DBEntry): Boolean {
        val title = dbEntry.title
        val description = dbEntry.description
        val imageUri = dbEntry.imageUri
        val favorite = dbEntry.favorite
        // Get the data repository in write mode
        val db = dbHelper.writableDatabase

        try {
            // Create a new map of values, where column names are the keys
            val values = ContentValues().apply {
                put(FeedContract.FeedEntry.COLUMN_DATE, date.toString())
                put(FeedContract.FeedEntry.COLUMN_TITLE, title)
                put(FeedContract.FeedEntry.COLUMN_DESCRIPTION, description)
                put(FeedContract.FeedEntry.COLUMN_IMAGE, imageUri)
                put(FeedContract.FeedEntry.COLUMN_FAVORITE, favorite) // not strictly necessary since always false by default
            }

            // Insert the new row, returning the primary key value of the new row
            db?.insert(FeedContract.FeedEntry.TABLE_FEED, null, values)
        } catch (exception: Exception) {
            db.close()
            return false
        } finally {
            db.close()
        }
        return true
    }

    fun setNewEntry(date: LocalDate, title: String, description: String, imageUri: String): Boolean {
        return setNewEntry(date, DBEntry(title, description, imageUri))
    }

    private fun changeFavorite(date: LocalDate, favorite: Boolean): Boolean {
        // Get the data repository in write mode
        val db = dbHelper.writableDatabase

        val values = ContentValues().apply {
            put(FeedContract.FeedEntry.COLUMN_FAVORITE, favorite)
        }

        // Which row to update, based on the date
        val selection = "${FeedContract.FeedEntry.COLUMN_DATE} LIKE ?"
        val selectionArgs = arrayOf(date.toString())
        val count = db.update(
            FeedContract.FeedEntry.TABLE_FEED,
            values,
            selection,
            selectionArgs)

        db.close()
        return (count == 1)
    }

    fun setFavorite(date: LocalDate): Boolean {
        return changeFavorite(date, true)
    }

    fun unsetFavorite(date: LocalDate): Boolean {
        return changeFavorite(date, false)
    }

    fun getFavorites(): List<Pair<LocalDate, String>> {
        val db = dbHelper.readableDatabase

        val projection =
            arrayOf(FeedContract.FeedEntry.COLUMN_DATE, FeedContract.FeedEntry.COLUMN_IMAGE)
        val selection = "${FeedContract.FeedEntry.COLUMN_FAVORITE} LIKE ?"
        val selectionArgs = arrayOf("1")
        val ordering = "${FeedContract.FeedEntry.COLUMN_DATE} DESC"

        val cursor = db.query(
            FeedContract.FeedEntry.TABLE_FEED,
            projection,
            selection,
            selectionArgs,
            null,
            null,
            ordering
        )

        val favoriteDatesImages = mutableListOf<Pair<LocalDate, String>>()

        with(cursor) {
            while (moveToNext()) {
                val date =
                    LocalDate.parse(getString(getColumnIndexOrThrow(FeedContract.FeedEntry.COLUMN_DATE)))

                val image = getString(getColumnIndexOrThrow(FeedContract.FeedEntry.COLUMN_IMAGE))

                if (image != "") {
                    favoriteDatesImages.add(Pair(date, image))
                }
            }
        }
        cursor.close()
        db.close()
        return favoriteDatesImages
    }

    fun getFeed(): List<Pair<Pair<LocalDate, String>, Pair<String, String>>> {
        // Get the data repository in read mode
        val db = dbHelper.readableDatabase

        val projection = arrayOf(FeedContract.FeedEntry.COLUMN_DATE, FeedContract.FeedEntry.COLUMN_TITLE, FeedContract.FeedEntry.COLUMN_DESCRIPTION, FeedContract.FeedEntry.COLUMN_IMAGE)
        val selection = "${FeedContract.FeedEntry.COLUMN_IMAGE} <> ''"
        val ordering = "${FeedContract.FeedEntry.COLUMN_DATE} DESC"

        val cursor = db.query(
            FeedContract.FeedEntry.TABLE_FEED,
            projection,
            selection,
            null,
            null,
            null,
            ordering
        )

        val datesTitlesDescriptionsImages = mutableListOf<Pair<Pair<LocalDate, String>, Pair<String, String>>>()
        with (cursor) {
            while (moveToNext()) {
                val date = LocalDate.parse(getString(getColumnIndexOrThrow(FeedContract.FeedEntry.COLUMN_DATE)))
                val title = getString(getColumnIndexOrThrow(FeedContract.FeedEntry.COLUMN_TITLE))
                val description = getString(getColumnIndexOrThrow(FeedContract.FeedEntry.COLUMN_DESCRIPTION))
                val image = getString(getColumnIndexOrThrow(FeedContract.FeedEntry.COLUMN_IMAGE))
                if (image != "") {
                    datesTitlesDescriptionsImages.add(Pair(Pair(date, title), Pair(description, image)))
                }
            }
        }
        cursor.close()
        db.close()
        return datesTitlesDescriptionsImages
    }

    fun changeImage(date: LocalDate, imageUri: String): Boolean {
        // Get the data repository in write mode
        val db = dbHelper.writableDatabase

        val values = ContentValues().apply {
            put(FeedContract.FeedEntry.COLUMN_IMAGE, imageUri)
        }

        // Which row to update, based on the date
        val selection = "${FeedContract.FeedEntry.COLUMN_DATE} LIKE ?"
        val selectionArgs = arrayOf(date.toString())
        val count = db.update(
            FeedContract.FeedEntry.TABLE_FEED,
            values,
            selection,
            selectionArgs)

        db.close()
        return (count == 1)
    }

    fun changeTitleDescription(date: LocalDate, title: String, description: String): Boolean {
        // Get the data repository in write mode
        val db = dbHelper.writableDatabase

        val values = ContentValues().apply {
            put(FeedContract.FeedEntry.COLUMN_TITLE, title)
            put(FeedContract.FeedEntry.COLUMN_DESCRIPTION, description)
        }

        // Which row to update, based on the date
        val selection = "${FeedContract.FeedEntry.COLUMN_DATE} LIKE ?"
        val selectionArgs = arrayOf(date.toString())
        val count = db.update(
            FeedContract.FeedEntry.TABLE_FEED,
            values,
            selection,
            selectionArgs)

        if (count == 0 && (title != "" || description != "")) {
            return setNewEntry(date, title, description, "")
        }

        db.close()
        return (count == 1)
    }


    fun entryExists(date: LocalDate): Boolean {
        // Get the data repository in read mode
        val db = dbHelper.readableDatabase

        val selection = "${FeedContract.FeedEntry.COLUMN_DATE} LIKE ?"
        val selectionArgs = arrayOf(date.toString())

        val cursor = db.query(
            FeedContract.FeedEntry.TABLE_FEED,
            null,
            selection,
            selectionArgs,
            null,
            null,
            null
        )

        val count = cursor.count
        cursor.close()
        db.close()
        return count == 1
    }

    private fun getEntry(date: LocalDate): DBEntry {
        // Get the data repository in read mode
        val db = dbHelper.readableDatabase

        val selection = "${FeedContract.FeedEntry.COLUMN_DATE} LIKE ?"
        val selectionArgs = arrayOf(date.toString())

        val cursor = db.query(
            FeedContract.FeedEntry.TABLE_FEED,
            null,
            selection,
            selectionArgs,
            null,
            null,
            null
        )

        when (cursor.count) {
            0 -> {
                cursor.close()
                db.close()
                throw NoSuchElementException()
            }
            1 -> {
                cursor.moveToNext()
                val title = cursor.getString(cursor.getColumnIndexOrThrow(FeedContract.FeedEntry.COLUMN_TITLE))
                val description = cursor.getString(cursor.getColumnIndexOrThrow(FeedContract.FeedEntry.COLUMN_DESCRIPTION))
                val imageUri = cursor.getString(cursor.getColumnIndexOrThrow(FeedContract.FeedEntry.COLUMN_IMAGE))
                val favorite = cursor.getInt(cursor.getColumnIndexOrThrow(FeedContract.FeedEntry.COLUMN_FAVORITE)) == 1
                cursor.close()
                db.close()
                return DBEntry(title, description, imageUri, favorite)
            }
            else -> {
                cursor.close()
                db.close()
                throw IllegalStateException()
            }
        }
    }

    fun getTitle(date: LocalDate): String {
        return getEntry(date).title
    }

    fun getDescription(date: LocalDate): String {
        return getEntry(date).description
    }

    fun getImage(date: LocalDate): String {
        return getEntry(date).imageUri
    }

    fun isFavorite(date: LocalDate): Boolean {
        return getEntry(date).favorite
    }

    fun deleteEntry(date: LocalDate): Boolean {
        // Get the data repository in write mode
        val db = dbHelper.writableDatabase

        // Which row to delete, based on the date
        val selection = "${FeedContract.FeedEntry.COLUMN_DATE} LIKE ?"
        val selectionArgs = arrayOf(date.toString())
        val deletedRows = db.delete(FeedContract.FeedEntry.TABLE_FEED, selection, selectionArgs)

        db.close()
        return (deletedRows == 1)
    }

}