package ch.mydoli.focal.ui.settings

import android.Manifest
import android.content.Context
import android.content.pm.PackageManager
import android.os.Bundle
import android.os.Environment
import android.util.Log
import android.view.View
import android.widget.Toast
import androidx.activity.result.contract.ActivityResultContracts
import androidx.appcompat.app.AppCompatDelegate
import androidx.core.content.ContextCompat
import androidx.core.net.toUri
import androidx.documentfile.provider.DocumentFile
import androidx.navigation.fragment.findNavController
import androidx.preference.ListPreference
import androidx.preference.Preference
import androidx.preference.PreferenceFragmentCompat
import androidx.preference.SwitchPreference
import ch.mydoli.focal.R
import ch.mydoli.focal.sql.database.FeedContract
import com.google.android.material.bottomnavigation.BottomNavigationView
import com.google.android.material.dialog.MaterialAlertDialogBuilder
import com.prolificinteractive.materialcalendarview.MaterialCalendarView
import org.threeten.bp.DayOfWeek
import java.io.File
import java.io.FileInputStream
import java.io.FileOutputStream
import java.time.LocalDate
import java.time.LocalTime
import java.time.format.DateTimeFormatter
import java.util.zip.ZipEntry
import java.util.zip.ZipInputStream
import java.util.zip.ZipOutputStream


class SettingsFragment : PreferenceFragmentCompat() {

    private lateinit var themePreference: ListPreference
    private lateinit var cameraImports: SwitchPreference

    private val requestCameraPermissionLauncher = registerForActivityResult(ActivityResultContracts.RequestPermission()) { granted ->
        cameraImports.isChecked = granted
    }

    private val requestStorageWriteLauncher = registerForActivityResult(ActivityResultContracts.OpenDocumentTree()) { uri ->
        if (uri != null) {
            // Can uncomment the following two lines if we need persistent access (i.e. across reboots) to this URI
            // val takeFlags: Int = Intent.FLAG_GRANT_READ_URI_PERMISSION or Intent.FLAG_GRANT_WRITE_URI_PERMISSION
            // requireContext().contentResolver.takePersistableUriPermission(uri, takeFlags)
            try {
                // Prepare necessary locations
                val appDataPath = "/data/${requireContext().packageName}"
                val dataDirectory = Environment.getDataDirectory()

                // Save existing database Path
                val dbPath = "$appDataPath/databases/${FeedContract.FeedDbHelper.DATABASE_NAME}"
                val internalDbPath = File(dataDirectory, dbPath)

                // Save existing photos path
                val photosPath = "${appDataPath}/files"
                val filesDirectory = File(dataDirectory, photosPath).listFiles()!!.asList()

                // Get current date and time
                val nowDate = LocalDate.now().format(DateTimeFormatter.ofPattern("yyyyMMdd"))
                val nowTime = LocalTime.now().format(DateTimeFormatter.ofPattern("HHmmss"))

                // Configure ZIP archive
                val externalDirectory = DocumentFile.fromTreeUri(requireContext(), uri)
                val exportedDb = "focal-$nowDate-$nowTime.zip"
                val zipFile = externalDirectory!!.createFile("application/zip", exportedDb)
                val zipFileDescriptor = requireContext().contentResolver.openFileDescriptor(zipFile!!.uri, "w")
                val exportedDbPath = zipFileDescriptor!!.fileDescriptor

                // Copy database and files over to ZIP archive
                val files = listOf(internalDbPath) + filesDirectory
                val zipOutputStream = ZipOutputStream(FileOutputStream(exportedDbPath))
                Toast.makeText(context, getString(R.string.exporting), Toast.LENGTH_LONG).show()
                for (file in files) {
                    val zipEntry = ZipEntry(file.parentFile!!.toUri().lastPathSegment + "/" + file.name)
                    zipOutputStream.putNextEntry(zipEntry)
                    val fileInputStream = file.inputStream()
                    val buffer = ByteArray(1024)
                    var length: Int
                    while (fileInputStream.read(buffer).also { length = it } > 0) {
                        zipOutputStream.write(buffer, 0, length)
                    }
                    fileInputStream.close()
                    zipOutputStream.closeEntry()
                }
                zipOutputStream.close()
                Toast.makeText(context, getString(R.string.export_completed), Toast.LENGTH_SHORT).show()
            } catch (exception: Exception) {
                Toast.makeText(context,"Error in export $exception", Toast.LENGTH_LONG).show()
                Log.e("File export error >>>>", exception.toString())
                exception.printStackTrace()
            }
        }
    }

    private val requestStorageReadLauncher = registerForActivityResult(ActivityResultContracts.OpenDocument()) { uri ->
        if (uri != null) {
            try {
                // Process uri and zip file
                val zipFileDescriptor = requireContext().contentResolver.openFileDescriptor(uri, "r")
                val importedDbPath = zipFileDescriptor!!.fileDescriptor
                val zipInputStream = ZipInputStream(FileInputStream(importedDbPath))

                // Prepare necessary locations
                val appDataPath = requireActivity().dataDir.toString()
                val dataDirectory = Environment.getDataDirectory()

                // Save existing database Path
                val dbPath = "$appDataPath/databases/${FeedContract.FeedDbHelper.DATABASE_NAME}"
                val internalDbPath = File(dataDirectory, dbPath)

                // Save existing photos path
                val photosPath = "${appDataPath}/files"
                val filesDirectory = File(dataDirectory, photosPath)

                // Delete current database and files (to be replaced)
                Toast.makeText(context, getString(R.string.importing), Toast.LENGTH_SHORT).show()
                deleteAll()

                // Iterate over all contents of the ZIP archive
                zipInputStream.use {
                    generateSequence { zipInputStream.nextEntry }
                        .filterNot { it.isDirectory }
                        .forEach { zipEntry ->
                            val filePath = zipEntry.name
                            val fileName = filePath.toUri().lastPathSegment
                            val fileParent = filePath.toUri().pathSegments.first()

                            val dbName = FeedContract.FeedDbHelper.DATABASE_NAME
                            val dbParent = requireActivity().getDatabasePath(dbName).parent!!.toUri().lastPathSegment!!

                            val copyPath = if (fileParent == dbParent && fileName == dbName) {
                                // We are handling the database itself
                                dbPath
                            } else {
                                // We are handling a picture or thumbnail
                                "$photosPath/$fileName"
                            }

                            // Copy the files over from ZIP
                            val buffer = ByteArray(1024)
                            var length: Int
                            val file = File(copyPath)
                            if (!file.exists()) {
                                file.createNewFile()
                            }
                            val fileOutputStream = FileOutputStream(copyPath)
                            while (zipInputStream.read(buffer).also { length = it } > 0) {
                                fileOutputStream.write(buffer, 0, length)
                            }
                            fileOutputStream.close()
                        }
                }
                zipInputStream.close()
                Toast.makeText(context, getString(R.string.import_completed), Toast.LENGTH_SHORT).show()
            } catch (exception: Exception) {
                Toast.makeText(context,"Error in import $exception", Toast.LENGTH_LONG).show()
                Log.e("File import error >>>>", exception.toString())
                exception.printStackTrace()
            }
        }
    }

    override fun onCreatePreferences(savedInstanceState: Bundle?, rootKey: String?) {
        setPreferencesFromResource(R.xml.fragment_settings, rootKey)

        // Select Light/Dark/System Mode
        themePreference = findPreference(getString(R.string.key_theme))!!
        themePreference.setOnPreferenceChangeListener { _, key ->
            setTheme(requireContext(), key as String)
        }

        // Choose start of the week
        // see setFirstDay()

        // Choose date format
        // see getDateFormat()

        // Select which kind of compression should be applied (default: 50% to minimize the attachment folder size)
        // see getQuality()

        // Choose whether direct camera imports should be possible
        cameraImports = findPreference(getString(R.string.key_camera))!!
        cameraImports.setOnPreferenceChangeListener { _, key ->
            when (key as Boolean) {
                true -> {
                    // Enable camera imports
                    // Ask for camera permissions
                    if (ContextCompat.checkSelfPermission(requireActivity(), Manifest.permission.CAMERA) != PackageManager.PERMISSION_GRANTED) {
                        requestCameraPermissionLauncher.launch(Manifest.permission.CAMERA)
                    }
                    true
                }
                false -> {
                    // Disable camera imports
                    cameraImports.isChecked = false
                    true
                }
            }
        }

        // Export database
        val exportPreference = findPreference<Preference>(getString(R.string.key_export))
        exportPreference?.setOnPreferenceClickListener {
            // Ask where to save DB to
            // TODO: Possibly set password for database encryption (if desired)
            requestStorageWriteLauncher.launch(null)
            true
        }

        // Import database
        val importPreference = findPreference<Preference>(getString(R.string.key_import))
        importPreference?.setOnPreferenceClickListener {
            // Ask from where to get DB
            // TODO: Possibly ask password for database decryption (if encrypted)
            MaterialAlertDialogBuilder(requireContext())
                .setMessage(R.string.warning_replace_all)
                .setPositiveButton(R.string.dialog_delete_cancel) { _, _ -> }
                .setNegativeButton(R.string.dialog_replace_proceed) { _, _ ->
                    requestStorageReadLauncher.launch(arrayOf("application/zip"))
                }.show()
            true
        }

        // Offer possibility to delete all images (with dialog to make sure this is desired)
        val deletePreference = findPreference<Preference>(getString(R.string.key_delete_all))
        deletePreference?.setOnPreferenceClickListener {
            MaterialAlertDialogBuilder(requireContext())
                .setMessage(R.string.warning_delete_all)
                .setPositiveButton(R.string.dialog_delete_cancel) { _, _ -> }
                .setNegativeButton(R.string.dialog_delete_proceed) { _, _ ->
                    Toast.makeText(context, getString(R.string.deleting), Toast.LENGTH_SHORT).show()
                    if (deleteAll()) {
                        Toast.makeText(context, getString(R.string.delete_completed), Toast.LENGTH_SHORT).show()
                    }
                }.show()
            true
        }

        // TODO: Possibly export to PDF/JPG with monthly view
        // TODO: Possibly export to diary in PDF format with small images (maybe with font customization)

        val aboutPreference = findPreference<Preference>(getString(R.string.key_about))
        aboutPreference?.setOnPreferenceClickListener {
            findNavController().navigate(R.id.action_navigation_settings_to_about, null, null, null)
            true
        }

        val licensesPreference = findPreference<Preference>(getString(R.string.key_licenses))
        licensesPreference?.setOnPreferenceClickListener {
            findNavController().navigate(R.id.action_navigation_settings_to_license, null, null, null)
            true
        }
    }

    override fun onResume() {
        super.onResume()
        // Show bottom navigation bar
        val navView: BottomNavigationView = requireActivity().findViewById(R.id.nav_view)
        navView.visibility = View.VISIBLE
    }

    private fun deleteAll(): Boolean {
        // Prepare necessary locations
        val appDataPath = "/data/${requireContext().packageName}"
        val dataDirectory = Environment.getDataDirectory()

        // Save existing database Path
        val dbPath = "$appDataPath/databases/${FeedContract.FeedDbHelper.DATABASE_NAME}"
        val internalDbPath = File(dataDirectory, dbPath)

        // Save existing photos path
        val photosPath = "${appDataPath}/files"
        val filesDirectory = File(dataDirectory, photosPath).listFiles()!!.asList()

        var deletedAll = false
        try {
            // Delete database
            deletedAll = internalDbPath.delete()

            // Delete all photos
            filesDirectory.forEach {
                deletedAll = it.delete() && deletedAll
            }
        } catch (exception: Exception) {
            Toast.makeText(context,"Error in deletion $exception", Toast.LENGTH_LONG).show()
            Log.e("Database delete error >>>>", exception.toString())
            exception.printStackTrace()
        }
        return deletedAll
    }

    companion object {
        private const val QUALITY_LESS: Int = 25
        private const val QUALITY_COMPROMISE: Int = 50
        private const val QUALITY_MORE: Int = 25

        fun setTheme(context: Context, key: String?): Boolean {
            when (key) {
                context.getString(R.string.system) -> {
                    // Set SYSTEM Theme
                    AppCompatDelegate.setDefaultNightMode(AppCompatDelegate.MODE_NIGHT_FOLLOW_SYSTEM)
                    return true
                }
                context.getString(R.string.light) -> {
                    // Set LIGHT Theme
                    AppCompatDelegate.setDefaultNightMode(AppCompatDelegate.MODE_NIGHT_NO)
                    return true
                }
                context.getString(R.string.dark) -> {
                    // Set DARK Theme
                    AppCompatDelegate.setDefaultNightMode(AppCompatDelegate.MODE_NIGHT_YES)
                    return true
                }
                else -> return false
            }
        }

        fun setFirstDay(context: Context, mcv: MaterialCalendarView, key: String?): Boolean {
            when (key) {
                context.getString(R.string.monday) -> {
                    // Set MONDAY
                    mcv.state().edit().setFirstDayOfWeek(DayOfWeek.MONDAY).commit()
                    return true
                }
                context.getString(R.string.sunday) -> {
                    // Set SUNDAY
                    mcv.state().edit().setFirstDayOfWeek(DayOfWeek.SUNDAY).commit()
                    return true
                }
                else -> return false
            }

        }

        fun getDateFormat(context: Context, date: LocalDate, key: String?): String {
            // Need to specify time to avoid conversion errors
            val time = date.atTime(0,0,0,0)

            return when (key) {
                context.getString(R.string.yyyymmdd) -> {
                    // Set yyyy-mm-dd
                    time.format(DateTimeFormatter.ofPattern("yyyy-MM-dd"))
                }
                context.getString(R.string.ddmmyyyy) -> {
                    // Set dd.mm.yyyy
                    time.format(DateTimeFormatter.ofPattern("dd.MM.yyyy"))
                }
                context.getString(R.string.dmmmmyyyy) -> {
                    // Set d mmmm yyyy
                    time.format(DateTimeFormatter.ofPattern("d MMMM yyyy"))
                }
                context.getString(R.string.mmmmdyyyy) -> {
                    // Set mmmm d, yyyy
                    time.format(DateTimeFormatter.ofPattern("MMMM d, yyyy"))
                }
                else -> {
                    // Set yyyy-mm-dd
                    date.toString()
                }
            }
        }

        fun getQuality(context: Context, key: String?): Int {
            return when (key) {
                context.getString(R.string.quality_less) -> {
                    // Set 25%
                    QUALITY_LESS
                }

                context.getString(R.string.quality_compromise) -> {
                    // Set 50%
                    QUALITY_COMPROMISE
                }

                context.getString(R.string.quality_more) -> {
                    // Set 75%
                    QUALITY_MORE
                }

                else -> QUALITY_COMPROMISE
            }
        }
    }
}